﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.PluginInfrastructure;
using HeuristicLab.Problems.Instances;

namespace HeuristicLab.Problems.Orienteering {

  [Item("Orienteering Problem", "Represents a single objective Orienteering Problem.")]
  [Creatable("Problems")]
  [StorableClass]
  public class OrienteeringProblem
    : SingleObjectiveHeuristicOptimizationProblem<IOrienteeringEvaluator, IIntegerVectorCreator>,
    IStorableContent, IProblemInstanceConsumer<CVRPData> {

    public string Filename { get; set; }

    #region Parameter Properties
    public OptionalValueParameter<DoubleMatrix> CoordinatesParameter {
      get { return (OptionalValueParameter<DoubleMatrix>)Parameters["Coordinates"]; }
    }
    public ValueParameter<DistanceMatrix> DistanceMatrixParameter {
      get { return (ValueParameter<DistanceMatrix>)Parameters["DistanceMatrix"]; }
    }

    public ValueParameter<IntValue> StartingPointParameter {
      get { return (ValueParameter<IntValue>)Parameters["StartingPoint"]; }
    }
    public ValueParameter<IntValue> TerminusPointParameter {
      get { return (ValueParameter<IntValue>)Parameters["TerminusPoint"]; }
    }
    public ValueParameter<DoubleValue> MaximumDistanceParameter {
      get { return (ValueParameter<DoubleValue>)Parameters["MaximumDistance"]; }
    }
    public ValueParameter<DoubleArray> ScoresParameter {
      get { return (ValueParameter<DoubleArray>)Parameters["Scores"]; }
    }
    public ValueParameter<DoubleValue> FixedPenaltyParameter {
      get { return (ValueParameter<DoubleValue>)Parameters["FixedPenalty"]; }
    }

    public OptionalValueParameter<IntegerVector> BestKnownSolutionParameter {
      get { return (OptionalValueParameter<IntegerVector>)Parameters["BestKnownSolution"]; }
    }
    #endregion

    #region Properties
    public DoubleMatrix Coordinates {
      get { return CoordinatesParameter.Value; }
      set { CoordinatesParameter.Value = value; }
    }
    public DistanceMatrix DistanceMatrix {
      get { return DistanceMatrixParameter.Value; }
      set { DistanceMatrixParameter.Value = value; }
    }
    public IntValue StartingPoint {
      get { return StartingPointParameter.Value; }
      set { StartingPointParameter.Value = value; }
    }
    public IntValue TerminusPoint {
      get { return TerminusPointParameter.Value; }
      set { TerminusPointParameter.Value = value; }
    }
    public DoubleValue MaximumDistance {
      get { return MaximumDistanceParameter.Value; }
      set { MaximumDistanceParameter.Value = value; }
    }
    public DoubleArray Scores {
      get { return ScoresParameter.Value; }
      set { ScoresParameter.Value = value; }
    }
    public DoubleValue FixedPenalty {
      get { return FixedPenaltyParameter.Value; }
      set { FixedPenaltyParameter.Value = value; }
    }
    public IntegerVector BestKnownSolution {
      get { return BestKnownSolutionParameter.Value; }
      set { BestKnownSolutionParameter.Value = value; }
    }
    private BestOrienteeringSolutionAnalyser BestOrienteeringSolutionAnalyser {
      get { return Operators.OfType<BestOrienteeringSolutionAnalyser>().SingleOrDefault(); }
    }
    #endregion

    [StorableConstructor]
    private OrienteeringProblem(bool deserializing)
      : base(deserializing) {
    }
    private OrienteeringProblem(OrienteeringProblem original, Cloner cloner)
      : base(original, cloner) {
      RegisterEventHandlers();
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new OrienteeringProblem(this, cloner);
    }
    public OrienteeringProblem()
      : base(new OrienteeringEvaluator(), new GreedyOrienteeringTourCreator()) {
      Parameters.Add(new OptionalValueParameter<DoubleMatrix>("Coordinates", "The x- and y-Coordinates of the points."));
      Parameters.Add(new ValueParameter<DistanceMatrix>("DistanceMatrix", "The matrix which contains the distances between the points."));
      Parameters.Add(new ValueParameter<IntValue>("StartingPoint", "Index of the starting point.", new IntValue(0)));
      Parameters.Add(new ValueParameter<IntValue>("TerminusPoint", "Index of the ending point.", new IntValue(0)));
      Parameters.Add(new ValueParameter<DoubleValue>("MaximumDistance", "The maximum distance constraint for a Orienteering solution."));
      Parameters.Add(new ValueParameter<DoubleArray>("Scores", "The scores of the points.", new DoubleArray()));
      Parameters.Add(new ValueParameter<DoubleValue>("FixedPenalty", "The penalty for each visited vertex."));
      Parameters.Add(new OptionalValueParameter<IntegerVector>("BestKnownSolution", "The best known solution of this Orienteering instance."));

      Maximization.Value = true;
      MaximizationParameter.Hidden = true;

      SolutionCreator.IntegerVectorParameter.ActualName = "OrienteeringSolution";

      InitializeInitialOrienteeringInstance();

      ParameterizeSolutionCreator();
      ParameterizeEvaluator();

      InitializeOperators();
      RegisterEventHandlers();
    }

    #region Events
    protected override void OnSolutionCreatorChanged() {
      base.OnSolutionCreatorChanged();
      SolutionCreator.IntegerVectorParameter.ActualNameChanged += SolutionCreator_IntegerVectorParameter_ActualNameChanged;
      ParameterizeSolutionCreator();
      ParameterizeEvaluator();
      ParameterizeAnalyzer();
      ParameterizeOperators();
    }
    protected override void OnEvaluatorChanged() {
      base.OnEvaluatorChanged();
      ParameterizeEvaluator();
      ParameterizeAnalyzer();
    }
    private void SolutionCreator_IntegerVectorParameter_ActualNameChanged(object sender, EventArgs e) {
      ParameterizeEvaluator();
      ParameterizeAnalyzer();
      ParameterizeOperators();
    }
    private void CoordinatesParameter_ValueChanged(object sender, EventArgs e) {
      if (Coordinates != null) {
        Coordinates.ItemChanged += new EventHandler<EventArgs<int, int>>(CoordinatesValue_ItemChanged);
        Coordinates.Reset += new EventHandler(CoordinatesValue_Reset);
      }
      ParameterizeSolutionCreator();
      CalculateDistanceMatrix();
    }
    private void CoordinatesValue_ItemChanged(object sender, EventArgs<int, int> e) {
      CalculateDistanceMatrix();
    }
    private void CoordinatesValue_Reset(object sender, EventArgs e) {
      ParameterizeSolutionCreator();
      CalculateDistanceMatrix();
    }
    private void StartingPointParameter_ValueChanged(object sender, EventArgs e) {
      ParameterizeEvaluator();
      ParameterizeAnalyzer();
    }

    private void TerminusPointParameter_ValueChanged(object sender, EventArgs e) {
      ParameterizeEvaluator();
      ParameterizeAnalyzer();
    }
    private void MaximumDistanceParameter_ValueChanged(object sender, EventArgs e) {
      ParameterizeEvaluator();
      ParameterizeAnalyzer();
    }
    private void ScoresParameter_ValueChanged(object sender, EventArgs e) {
      ParameterizeEvaluator();
      ParameterizeAnalyzer();
      ParameterizeSolutionCreator();

      ScoresParameter.Value.Reset += new EventHandler(ScoresValue_Reset);
    }
    private void ScoresValue_Reset(object sender, EventArgs e) {
      ParameterizeSolutionCreator();
    }
    private void FixedPenaltyParameter_ValueChanged(object sender, EventArgs e) {
      ParameterizeEvaluator();
      ParameterizeAnalyzer();
    }
    #endregion

    #region Helpers
    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      RegisterEventHandlers();
    }

    private void RegisterEventHandlers() {
      SolutionCreator.IntegerVectorParameter.ActualNameChanged += SolutionCreator_IntegerVectorParameter_ActualNameChanged;

      CoordinatesParameter.ValueChanged += CoordinatesParameter_ValueChanged;
      if (CoordinatesParameter.Value != null) {
        CoordinatesParameter.Value.ItemChanged += CoordinatesValue_ItemChanged;
        CoordinatesParameter.Value.Reset += CoordinatesValue_Reset;
      }

      StartingPointParameter.ValueChanged += StartingPointParameter_ValueChanged;
      TerminusPointParameter.ValueChanged += TerminusPointParameter_ValueChanged;
      MaximumDistanceParameter.ValueChanged += MaximumDistanceParameter_ValueChanged;

      ScoresParameter.ValueChanged += ScoresParameter_ValueChanged;
      ScoresParameter.Value.Reset += ScoresValue_Reset;
      FixedPenaltyParameter.ValueChanged += FixedPenaltyParameter_ValueChanged;
    }

    private void ParameterizeSolutionCreator() {
      if (SolutionCreator is GreedyOrienteeringTourCreator) {
        var creator = (GreedyOrienteeringTourCreator)SolutionCreator;
        creator.DistanceMatrixParameter.ActualName = DistanceMatrixParameter.Name;
        creator.ScoresParameter.ActualName = ScoresParameter.Name;
        creator.MaximumDistanceParameter.ActualName = MaximumDistanceParameter.Name;
        creator.StartingPointParameter.ActualName = StartingPointParameter.Name;
        creator.TerminusPointParameter.ActualName = TerminusPointParameter.Name;
        creator.FixedPenaltyParameter.ActualName = FixedPenaltyParameter.Name;
      }
    }
    private void ParameterizeEvaluator() {
      if (Evaluator is OrienteeringEvaluator) {
        var orienteeringEvaluator = (OrienteeringEvaluator)Evaluator;
        orienteeringEvaluator.SolutionParameter.ActualName = SolutionCreator.IntegerVectorParameter.ActualName;
        orienteeringEvaluator.ScoresParameter.ActualName = ScoresParameter.Name;
      }
    }
    private void ParameterizeAnalyzer() {
      if (BestOrienteeringSolutionAnalyser != null) {
        BestOrienteeringSolutionAnalyser.QualityParameter.ActualName = Evaluator.QualityParameter.ActualName;

        BestOrienteeringSolutionAnalyser.IntegerVector.ActualName = SolutionCreator.IntegerVectorParameter.ActualName;
        BestOrienteeringSolutionAnalyser.CoordinatesParameter.ActualName = CoordinatesParameter.Name;
        BestOrienteeringSolutionAnalyser.ScoresParameter.ActualName = ScoresParameter.Name;

        BestOrienteeringSolutionAnalyser.ResultsParameter.ActualName = "Results";
        BestOrienteeringSolutionAnalyser.BestKnownQualityParameter.ActualName = BestKnownQualityParameter.Name;
        BestOrienteeringSolutionAnalyser.BestKnownSolutionParameter.ActualName = BestKnownSolutionParameter.Name;
      }
    }
    private void InitializeOperators() {
      Operators.Add(new BestOrienteeringSolutionAnalyser());
      ParameterizeAnalyzer();

      var operators = ApplicationManager.Manager.GetInstances<IIntegerVectorOperator>();
      Operators.AddRange(operators);
      ParameterizeOperators();
    }
    private void ParameterizeOperators() {
      //foreach (var op in Operators.OfType<IIntegerVectorManipulator>())
      //  op.IntegerVectorParameter.ActualName = SolutionCreator.IntegerVectorParameter.ActualName;
      foreach (var op in Operators.OfType<IIntegerVectorMultiNeighborhoodShakingOperator>())
        op.IntegerVectorParameter.ActualName = SolutionCreator.IntegerVectorParameter.ActualName;
      foreach (var op in Operators.OfType<ISingleObjectiveImprovementOperator>())
        op.SolutionParameter.ActualName = SolutionCreator.IntegerVectorParameter.ActualName;
    }
    #endregion

    private void CalculateDistanceMatrix() {
      var distances = new double[Coordinates.Rows, Coordinates.Rows];
      for (int i = 0; i < Coordinates.Rows; i++) {
        for (int j = 0; j < Coordinates.Rows; j++) {
          double distanceX = Math.Abs(Coordinates[i, 0] - Coordinates[j, 0]);
          double distanceY = Math.Abs(Coordinates[i, 1] - Coordinates[j, 1]);
          distances[i, j] = Math.Sqrt(Math.Pow(distanceX, 2) + Math.Pow(distanceY, 2));
        }
      }
      DistanceMatrix = new DistanceMatrix(distances);
    }

    private void InitializeInitialOrienteeringInstance() {
      Coordinates = new DoubleMatrix(new double[21, 2] {
        {  4.60,  7.10 }, {  5.70, 11.40 }, {  4.40, 12.30 }, {  2.80, 14.30 }, {  3.20, 10.30 },
        {  3.50,  9.80 }, {  4.40,  8.40 }, {  7.80, 11.00 }, {  8.80,  9.80 }, {  7.70,  8.20 },
        {  6.30,  7.90 }, {  5.40,  8.20 }, {  5.80,  6.80 }, {  6.70,  5.80 }, { 13.80, 13.10 },
        { 14.10, 14.20 }, { 11.20, 13.60 }, {  9.70, 16.40 }, {  9.50, 18.80 }, {  4.70, 16.80 },
        {  5.00,  5.60 }
      });
      CalculateDistanceMatrix();

      StartingPoint.Value = 0;
      TerminusPoint.Value = 20;
      MaximumDistance.Value = 30;

      Scores = new DoubleArray(new double[21] { 0, 20, 20, 30, 15, 15, 10, 20, 20, 20, 15, 10, 10, 25, 40, 40, 30, 30, 50, 30, 0 });
    }

    public void Load(CVRPData data) {
      if (data.Coordinates == null && data.Distances == null)
        throw new System.IO.InvalidDataException("The given instance specifies neither coordinates nor distances!");
      if (data.Coordinates != null && data.Coordinates.GetLength(1) != 2)
        throw new System.IO.InvalidDataException("The coordinates of the given instance are not in the right format, there need to be one row for each customer and two columns for the x and y coordinates.");

      Name = data.Name;
      Description = data.Description;

      Coordinates = data.Coordinates != null ? new DoubleMatrix(data.Coordinates) : null;
      if (Coordinates == null)
        DistanceMatrix = new DistanceMatrix(data.Distances);
      StartingPoint = new IntValue(0);// Depot is interpreted as start and endpoint (default)
      TerminusPoint = new IntValue(0);

      MaximumDistance = new DoubleValue(data.Capacity); // capacity is interpreted as max distance
      Scores = new DoubleArray(data.Demands); // demands are interpreted as scores

      BestKnownQuality = null;
      BestKnownSolution = null;
    }
  }
}